package abc.meta.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Node;
import polyglot.ast.TypeNode;
import polyglot.types.Flags;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import abc.aspectj.ast.AspectBody;
import abc.aspectj.ast.PerClause;
import abc.aspectj.types.AspectType;
import abc.meta.weaving.aspectinfo.Aspect;
import abc.weaving.aspectinfo.AbcClass;
import abc.weaving.aspectinfo.AbcFactory;
import abc.weaving.aspectinfo.Formal;
import abc.weaving.aspectinfo.GlobalAspectInfo;
import abc.weaving.aspectinfo.MethodCategory;
import abc.weaving.aspectinfo.MethodSig;
import abc.weaving.aspectinfo.Per;

/**
 * This class overrides the original AspectDecl, providing an additional metalevel.
 * 
 * @author Eric Bodden
 */
public class AspectDecl_c extends abc.aspectj.ast.AspectDecl_c implements
        AspectDecl {

    /** The meta level tag of this aspect. */
    protected MetaTag metaTag;

    /**
     * Same as {@link abc.aspectj.ast.AspectDecl_c#AspectDecl_c(Position, boolean, Flags, String, TypeNode, List, PerClause, AspectBody)}
     * but with an additional meta level as parameter.
     * 
     * @param metaLevel The metalevel of this aspect.
     */
    public AspectDecl_c(Position pos, boolean is_privileged, Flags flags,
            String name, TypeNode superClass, List interfaces, PerClause per,
            AspectBody body, MetaTag metaTag) {
        super(pos, is_privileged, flags, name, superClass, interfaces, per,
                body);
        this.metaTag = metaTag;
    }

    /**
     * Here we enclude a check that the meta level is >= 0.
     * Also we check that super classes / aspects are of the correct meta level.
     */
    public Node typeCheck(TypeChecker tc) throws SemanticException {
        //check that the meta level is in the valid range 
        if (metaTag.getMetaLevel() < 0) {
            throw new SemanticException("Metalevel must be >= 0.", metaTag.position());
        }
        
        //check that super types are of a consistent meta level
        final Type superType = type().superType();
        final Type OBJECT = tc.typeSystem().typeForName("java.lang.Object");
        if(!superType.equals(OBJECT)) {
            //if we have a real super type (not "Object")
            if(superType instanceof abc.meta.types.AspectType) {
                //if we have a super aspect, the meta levels have to be the same
                abc.meta.types.AspectType superAspectType = (abc.meta.types.AspectType) superType;
                if(superAspectType.getMetaLevel()!=getMetaLevel()) {
                    throw new SemanticException("Super aspect type '"+superAspectType.toString()+ "' has " +
                            "meta level "+superAspectType.getMetaLevel()+" while this aspect has " +
                            "meta level "+getMetaLevel()+". The meta levels have to coincide.",
                            superClass().position());
                }                
            } else {
                //aspects are not allowed to extend classes, because classes have a different meta level
                throw new SemanticException("Aspects are not allowed to extend classes because this " +
                        "would imply an ambiguity in the meta level.", superClass().position());
            }
        }
        
        //delegate up
        return super.typeCheck(tc);
    }

    /**
     * {@inheritDoc}
     */
    public int getMetaLevel() {
        return metaTag.getMetaLevel();
    }

    /**
     * This is essentially the same as {@link abc.aspectj.ast.AspectDecl_c#update(GlobalAspectInfo, abc.weaving.aspectinfo.Aspect)}
     * except that we generate an aspect info which also holds the appropriate meta level.
     */
    //copied from abc.aspectj.ast.AspectDecl_c
    public void update(GlobalAspectInfo gai, abc.weaving.aspectinfo.Aspect current_aspect) {
        Per p = (per == null ? null : per.makeAIPer());
        AbcClass cl = AbcFactory.AbcClass(type());
        
        //MODIFIED HERE
        abc.weaving.aspectinfo.Aspect a = new Aspect(cl, p, getMetaLevel(), position());
        //END OF MODIFICATION
        
        gai.addAspect(a);

        List fl = new ArrayList();
        if (((AspectType) type()).perObject()) {
            fl.add(new Formal(AbcFactory.AbcType(soot.RefType
                    .v("java.lang.Object")), "obj", position()));
        }

        List el = new ArrayList();
        // FIXME: Do these methods declare any exceptions?
        if (!flags().isAbstract()) {

            MethodSig aspectOf = new MethodSig(soot.Modifier.PUBLIC
                    | soot.Modifier.STATIC, cl, AbcFactory.AbcType(type()),
                    "aspectOf", fl, el, position());
            MethodSig hasAspect = new MethodSig(soot.Modifier.PUBLIC
                    | soot.Modifier.STATIC, cl, AbcFactory
                    .AbcType(soot.BooleanType.v()), "hasAspect", fl, el,
                    position());

            MethodCategory.register(aspectOf, MethodCategory.ASPECT_INSTANCE);
            MethodCategory.register(hasAspect, MethodCategory.ASPECT_INSTANCE);
        }
    }

}
