package abc.meta.ast;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import polyglot.ast.Expr;
import polyglot.ast.Ext;
import polyglot.ast.JL;
import polyglot.ast.Node;
import polyglot.ast.Precedence;
import polyglot.ast.Stmt;
import polyglot.types.Context;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.util.CodeWriter;
import polyglot.util.Position;
import polyglot.visit.AddMemberVisitor;
import polyglot.visit.AmbiguityRemover;
import polyglot.visit.AscriptionVisitor;
import polyglot.visit.ExceptionChecker;
import polyglot.visit.NodeVisitor;
import polyglot.visit.PrettyPrinter;
import polyglot.visit.Translator;
import polyglot.visit.TypeBuilder;
import polyglot.visit.TypeChecker;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.ast.MakesAspectMethods;
import abc.aspectj.ast.PCAdviceExecution;
import abc.aspectj.ast.Pointcut_c;
import abc.aspectj.types.AJTypeSystem;
import abc.aspectj.visit.AspectMethods;
import abc.meta.weaving.aspectinfo.MetaTagPointcut;
import abc.weaving.aspectinfo.Pointcut;
import abc.weaving.aspectinfo.WithinAdvice;

/**
 * This class implements a decorator to basic pointcuts holding a meta level.
 * <br>
 * Actually, this class is just a decorator.
 * Unfortunately, it has to be a bit more fancy than usual because polyglot makes
 * nondestructive updates and so forth.
 * In particular, we cannot always just return the return value of the delegate
 * but rather have to return an updated version of ourselves, containing that return value.
 * @author Eric Bodden
 */
public class PCMetaTag_c extends Pointcut_c implements PCMetaTag, MakesAspectMethods {

    /** The decorated pointcut. */
    protected Pointcut_c delegate;
    
    /** The meta level of the decorated pointcut. */
    protected int metaLevel;
    
    /**
     * @param delegate the decorated pointcut
     * @param metaLevel the meta level of the decorated pointcut
     * @param pos the source code position
     */
    public PCMetaTag_c(Pointcut_c delegate, int metaLevel, Position pos) {
        super(pos);
        this.delegate = delegate;
        this.metaLevel = metaLevel;
    }

    /**
     * Here we return the appropriately decorated aspect info.
     * @see abc.aspectj.ast.Pointcut#makeAIPointcut()
     */
    public Pointcut makeAIPointcut() {
        if(delegate instanceof PCAdviceExecution) {
            //for adviceexecution pointcuts, we do not return
            //meta(i) (withinadvice() && execution())
            //but rather
            //(meta(i) withinadvice() && execution())
            //because this is semantically equivalent and allows us lateron to easily
            //discover the combination "meta(i) adviceexecution" again for matching
            return abc.weaving.aspectinfo.AndPointcut.construct
            (new MetaTagPointcut(new WithinAdvice(position()),metaLevel, position()),
             new abc.weaving.aspectinfo.Execution(position()),
             position());
        } else {
            //in the other cases just decorate the delegate's aspect info
            return new MetaTagPointcut(delegate.makeAIPointcut(), metaLevel, position());
        }        
    }


    /**
     * We have to make sure that during nondestructive updates of polyglot, we do not
     * return the delegate but rather again a meta pointcut with the updated delegate.
     * This is what this method provides.
     */
    protected PCMetaTag_c reconstruct(Node node) {
        Pointcut_c otherDelegate = (Pointcut_c) node;
        if(otherDelegate != this.delegate) {
            PCMetaTag_c n=(PCMetaTag_c) copy();
            n.delegate = otherDelegate;
            return n;
        }
        return this;
    }

    /**
     * {@inheritDoc}
     */
    public Position position() {
        //return not the position of the delegate but rather the one of the meta tag
        //(looks nicer in error messages)
        return super.position();
    }

    /**
     * {@inheritDoc}
     */
    public JL del() {
        //make sure to not return the delegate's delegate cause otherwise we would bypass ourselves 
        return super.del();
    }
    
    /**
     * {@inheritDoc}
     */
    public Object copy() {
        //make sure to copy ourselves and not the delegate
        return super.copy();
    }

    /**
     * {@inheritDoc}
     */
    public int getMetaLevel() {
        return metaLevel;
    }
    
    //////////////////
    //the following methods just delegate and reconstruct
    //////////////////
    
    /**
     * {@inheritDoc}
     */
    public Node ext(Ext ext) {
        return reconstruct(delegate.ext(ext));
    }

    /**
     * {@inheritDoc}
     */
    public Node typeCheck(TypeChecker tc) throws SemanticException {
        return reconstruct(delegate.typeCheck(tc));
    }

    /**
     * @param c
     * @see polyglot.ext.jl.ast.Node_c#addDecls(polyglot.types.Context)
     */
    public void addDecls(Context c) {
        delegate.addDecls(c);
    }

    /**
     * @param am
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#addMembers(polyglot.visit.AddMemberVisitor)
     */
    public Node addMembers(AddMemberVisitor am) throws SemanticException {
        return reconstruct(delegate.addMembers(am));
    }

    /**
     * @param am
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#addMembersEnter(polyglot.visit.AddMemberVisitor)
     */
    public NodeVisitor addMembersEnter(AddMemberVisitor am) throws SemanticException {
        return delegate.addMembersEnter(am);
    }

    /**
     * @param am
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#addMembersOverride(polyglot.visit.AddMemberVisitor)
     */
    public Node addMembersOverride(AddMemberVisitor am) throws SemanticException {
        return delegate.addMembersOverride(am);
    }

    /**
     * @param tb
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#buildTypes(polyglot.visit.TypeBuilder)
     */
    public Node buildTypes(TypeBuilder tb) throws SemanticException {
        return reconstruct(delegate.buildTypes(tb));
    }

    /**
     * @param tb
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#buildTypesEnter(polyglot.visit.TypeBuilder)
     */
    public NodeVisitor buildTypesEnter(TypeBuilder tb) throws SemanticException {
        return delegate.buildTypesEnter(tb);
    }

    /**
     * @param tb
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#buildTypesOverride(polyglot.visit.TypeBuilder)
     */
    public Node buildTypesOverride(TypeBuilder tb) throws SemanticException {
        return reconstruct(delegate.buildTypesOverride(tb));
    }

    /**
     * @param formals
     * @throws SemanticException
     * @see abc.aspectj.ast.Pointcut_c#checkFormals(java.util.List)
     */
    public void checkFormals(List formals) throws SemanticException {
        delegate.checkFormals(formals);
    }

    /**
     * @param child
     * @param av
     * @return
     * @see polyglot.ext.jl.ast.Node_c#childExpectedType(polyglot.ast.Expr, polyglot.visit.AscriptionVisitor)
     */
    public Type childExpectedType(Expr child, AscriptionVisitor av) {
        return delegate.childExpectedType(child, av);
    }

    /**
     * @param del
     * @return
     * @see polyglot.ext.jl.ast.Node_c#del(polyglot.ast.JL)
     */
    public Node del(JL del) {
        return reconstruct(delegate.del(del));
    }

    /**
     * @param ar
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#disambiguate(polyglot.visit.AmbiguityRemover)
     */
    public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
        return reconstruct(delegate.disambiguate(ar));
    }

    /**
     * @param ar
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#disambiguateEnter(polyglot.visit.AmbiguityRemover)
     */
    public NodeVisitor disambiguateEnter(AmbiguityRemover ar) throws SemanticException {
        return delegate.disambiguateEnter(ar);
    }

    /**
     * @param ar
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#disambiguateOverride(polyglot.visit.AmbiguityRemover)
     */
    public Node disambiguateOverride(AmbiguityRemover ar) throws SemanticException {
        return reconstruct(delegate.disambiguateOverride(ar));
    }

    /**
     * @param w
     * @see polyglot.ext.jl.ast.Node_c#dump(polyglot.util.CodeWriter)
     */
    public void dump(CodeWriter w) {
        delegate.dump(w);
    }

    /**
     * @param c
     * @return
     * @see polyglot.ext.jl.ast.Node_c#enterScope(polyglot.types.Context)
     */
    public Context enterScope(Context c) {
        return delegate.enterScope(c);
    }

    /**
     * @param child
     * @param c
     * @return
     * @see polyglot.ext.jl.ast.Node_c#enterScope(polyglot.ast.Node, polyglot.types.Context)
     */
    public Context enterScope(Node child, Context c) {
        return delegate.enterScope(child, c);
    }

    /**
     * @param obj
     * @return
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj) {
        return delegate.equals(obj);
    }

    /**
     * @param ec
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#exceptionCheck(polyglot.visit.ExceptionChecker)
     */
    public Node exceptionCheck(ExceptionChecker ec) throws SemanticException {
        return reconstruct(delegate.exceptionCheck(ec));
    }

    /**
     * @param ec
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#exceptionCheckEnter(polyglot.visit.ExceptionChecker)
     */
    public NodeVisitor exceptionCheckEnter(ExceptionChecker ec) throws SemanticException {
        return delegate.exceptionCheckEnter(ec);
    }

    /**
     * @param ec
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#exceptionCheckOverride(polyglot.visit.ExceptionChecker)
     */
    public Node exceptionCheckOverride(ExceptionChecker ec) throws SemanticException {
        return reconstruct(delegate.exceptionCheckOverride(ec));
    }

    /**
     * @return
     * @see polyglot.ext.jl.ast.Node_c#ext()
     */
    public Ext ext() {
        return delegate.ext();
    }

    /**
     * @param n
     * @param ext
     * @return
     * @see polyglot.ext.jl.ast.Node_c#ext(int, polyglot.ast.Ext)
     */
    public Node ext(int n, Ext ext) {
        return reconstruct(delegate.ext(n, ext));
    }

    /**
     * @param n
     * @return
     * @see polyglot.ext.jl.ast.Node_c#ext(int)
     */
    public Ext ext(int n) {
        return delegate.ext(n);
    }

    /**
     * @return
     * @see java.lang.Object#hashCode()
     */
    public int hashCode() {
        return delegate.hashCode();
    }

    /**
     * @param node
     * @see polyglot.ext.jl.ast.Node_c#init(polyglot.ast.Node)
     */
    public void init(Node node) {
        delegate.init(node);
    }

    /**
     * @return
     * @see abc.aspectj.ast.Pointcut#isDynamic()
     */
    public boolean isDynamic() {
        return delegate.isDynamic();
    }

    /**
     * @return
     * @throws SemanticException
     * @see abc.aspectj.ast.Pointcut_c#mayBind()
     */
    public Collection mayBind() throws SemanticException {
        return delegate.mayBind();
    }

    /**
     * @return
     * @see abc.aspectj.ast.Pointcut_c#mustBind()
     */
    public Collection mustBind() {
        return delegate.mustBind();
    }

    /**
     * @return
     * @see polyglot.ext.jl.ast.Node_c#node()
     */
    public Node node() {
        return reconstruct(delegate.node());
    }

    /**
     * @return
     * @see abc.aspectj.ast.Pointcut#pcRefs()
     */
    public Set pcRefs() {
        return delegate.pcRefs();
    }

    /**
     * @param position
     * @return
     * @see polyglot.ext.jl.ast.Node_c#position(polyglot.util.Position)
     */
    public Node position(Position position) {
        return reconstruct(delegate.position(position));
    }

    /**
     * @return
     * @see abc.aspectj.ast.Pointcut_c#precedence()
     */
    public Precedence precedence() {
        return delegate.precedence();
    }

    /**
     * @param w
     * @param pp
     * @see polyglot.ext.jl.ast.Node_c#prettyPrint(polyglot.util.CodeWriter, polyglot.visit.PrettyPrinter)
     */
    public void prettyPrint(CodeWriter w, PrettyPrinter pp) {
        w.write("meta(" + metaLevel + ") ");
        delegate.prettyPrint(w, pp);
    }

    /**
     * @param child
     * @param w
     * @param pp
     * @see polyglot.ext.jl.ast.Node_c#print(polyglot.ast.Node, polyglot.util.CodeWriter, polyglot.visit.PrettyPrinter)
     */
    public void print(Node child, CodeWriter w, PrettyPrinter pp) {
        delegate.print(child, w, pp);
    }

    /**
     * @param n
     * @param w
     * @param pp
     * @see polyglot.ext.jl.ast.Node_c#printBlock(polyglot.ast.Node, polyglot.util.CodeWriter, polyglot.visit.PrettyPrinter)
     */
    public void printBlock(Node n, CodeWriter w, PrettyPrinter pp) {
        delegate.printBlock(n, w, pp);
    }

    /**
     * @param pc
     * @param associative
     * @param w
     * @param pp
     * @see abc.aspectj.ast.Pointcut_c#printSubExpr(abc.aspectj.ast.Pointcut, boolean, polyglot.util.CodeWriter, polyglot.visit.PrettyPrinter)
     */
    public void printSubExpr(abc.aspectj.ast.Pointcut pc, boolean associative, CodeWriter w, PrettyPrinter pp) {
        delegate.printSubExpr(pc, associative, w, pp);
    }

    /**
     * @param stmt
     * @param w
     * @param pp
     * @see polyglot.ext.jl.ast.Node_c#printSubStmt(polyglot.ast.Stmt, polyglot.util.CodeWriter, polyglot.visit.PrettyPrinter)
     */
    public void printSubStmt(Stmt stmt, CodeWriter w, PrettyPrinter pp) {
        delegate.printSubStmt(stmt, w, pp);
    }

    /**
     * @param ts
     * @return
     * @see polyglot.ext.jl.ast.Node_c#throwTypes(polyglot.types.TypeSystem)
     */
    public List throwTypes(TypeSystem ts) {
        return delegate.throwTypes(ts);
    }

    /**
     * @return
     * @see polyglot.ext.jl.ast.Node_c#toString()
     */
    public String toString() {
        return "meta(" + metaLevel + ") " + delegate.toString();
    }

    /**
     * @param w
     * @param tr
     * @see polyglot.ext.jl.ast.Node_c#translate(polyglot.util.CodeWriter, polyglot.visit.Translator)
     */
    public void translate(CodeWriter w, Translator tr) {
        delegate.translate(w, tr);
    }

    /**
     * @param tc
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#typeCheckEnter(polyglot.visit.TypeChecker)
     */
    public NodeVisitor typeCheckEnter(TypeChecker tc) throws SemanticException {
        return delegate.typeCheckEnter(tc);
    }

    /**
     * @param tc
     * @return
     * @throws SemanticException
     * @see polyglot.ext.jl.ast.Node_c#typeCheckOverride(polyglot.visit.TypeChecker)
     */
    public Node typeCheckOverride(TypeChecker tc) throws SemanticException {
        return reconstruct(delegate.typeCheckOverride(tc));
    }

    /**
     * @param v
     * @return
     * @see polyglot.ext.jl.ast.Node_c#visit(polyglot.visit.NodeVisitor)
     */
    public Node visit(NodeVisitor v) {
        return reconstruct(delegate.visit(v));
    }

    /**
     * @param n
     * @param v
     * @return
     * @see polyglot.ext.jl.ast.Node_c#visitChild(polyglot.ast.Node, polyglot.visit.NodeVisitor)
     */
    public Node visitChild(Node n, NodeVisitor v) {
        return reconstruct(delegate.visitChild(n, v));
    }

    /**
     * @param v
     * @return
     * @see polyglot.ext.jl.ast.Node_c#visitChildren(polyglot.visit.NodeVisitor)
     */
    public Node visitChildren(NodeVisitor v) {
        return reconstruct(delegate.visitChildren(v));
    }

    /**
     * @param parent
     * @param v
     * @return
     * @see polyglot.ext.jl.ast.Node_c#visitEdge(polyglot.ast.Node, polyglot.visit.NodeVisitor)
     */
    public Node visitEdge(Node parent, NodeVisitor v) {
        return reconstruct(delegate.visitEdge(parent, v));
    }

    /**
     * @param l
     * @param v
     * @return
     * @see polyglot.ext.jl.ast.Node_c#visitList(java.util.List, polyglot.visit.NodeVisitor)
     */
    public List visitList(List l, NodeVisitor v) {
        return delegate.visitList(l, v);
    }
    
    //it comes even worse: we have to delegate even for all special instances of 
    //meta tags 

    /**
     * {@inheritDoc}
     */
    public void aspectMethodsEnter(AspectMethods visitor) {
        if(delegate instanceof MakesAspectMethods) {
            ((MakesAspectMethods) delegate).aspectMethodsEnter(visitor);
        }
    }

    /**
     * {@inheritDoc}
     */
    public Node aspectMethodsLeave(AspectMethods visitor, AJNodeFactory nf, AJTypeSystem ts) {
        if(delegate instanceof MakesAspectMethods) {
            return ((MakesAspectMethods) delegate).aspectMethodsLeave(visitor, nf, ts);
        } else {
            return this;
        }
    }

}
