package abc.meta.weaving.aspectinfo;

import java.util.Hashtable;
import java.util.Set;

import polyglot.types.ClassType;
import polyglot.types.SemanticException;
import polyglot.util.Position;
import soot.SootClass;
import abc.meta.MetaLevelException;
import abc.meta.extension.ExtDefaultMetaLevelAssign;
import abc.meta.types.AspectType;
import abc.weaving.aspectinfo.AbcFactory;
import abc.weaving.aspectinfo.Aspect;
import abc.weaving.aspectinfo.Pointcut;
import abc.weaving.aspectinfo.Unification;
import abc.weaving.aspectinfo.WithinAdvice;
import abc.weaving.matching.MatchingContext;
import abc.weaving.residues.NeverMatch;
import abc.weaving.residues.Residue;

/**
 * A decorator for pointcut aspect infos implementing type checkind and matching.
 * We have to do typechecking very late (just at inlining time of the pointcuts)
 * in order to handle named pointcuts correctly:
 * If a named pointcut is allowed in an aspect depends on the pointcuts it points to!
 * @author Eric Bodden
 */
public class MetaTagPointcut extends Pointcut {
    
    /** The decorated pointcut. */
    protected Pointcut delegate;
    
    /** The meta level. */
    protected int level;

    /**
     * Constructs a new aspect info for a meta pointcut decorator.
     * @param delegate the decorated pointcut
     * @param metaLevel the meta level
     * @param pos the source code position of the corresponding AST node
     */
    public MetaTagPointcut(Pointcut delegate, int metaLevel, Position pos) {
        super(pos);
        this.delegate = delegate;
        this.level = metaLevel;
    }
    
    /**
     * Implements the matching strategy for the meta pointcut.
     * A pointcut <code>meta(i) pc</code> matches a class <code>c</code>
     * if <code>pc</code> matches and if ...
     * <ul>
     * <li><code>c</code> is a class and <i>i</> is 0
     * <li><code>c</code> is an aspect with level <i>l</i> and
     *   <ul>
     *   <li><code>pc</code> is an <code>adviceexecution</code> pointcut and <i>i=l</i>
     *   <li><code>pc</code> is another type of pointcut and <i>i=l+1</i>
     *   </ul>
     * </ul>
     * For instance matches <code>meta(1) execution</code> on a <i>level-0</i> aspect as well as
     * <code>meta(0) adviceexecution</code> (= <code>adviceexecution</code>).
     */
    public Residue matchesAt(MatchingContext mc) throws SemanticException {
        SootClass cls = mc.getSootClass();
        
        ClassType ct = AbcFactory.sootClassToClassType(cls);
        if(ct instanceof AspectType) {
            //if we have an aspect
            AspectType type = (AspectType) ct;
            int targetLevel = type.getMetaLevel();
            
            int matchLevel;
            if(delegate instanceof WithinAdvice) {
                matchLevel = this.level;
            } else {
                matchLevel = this.level - 1;
            }
            //make sure that the levels coincide;
            if(matchLevel  != targetLevel) {
                return NeverMatch.v();
            }           
        } else {
            //we have a class; only DEFAULT level pointcuts apply here
            if(level!=ExtDefaultMetaLevelAssign.DEFAULT_LEVEL_STANDARD) {
                return NeverMatch.v();
            }           
        }

        return delegate.matchesAt(mc);
    }

    /**
     * Implements a type check during inlining, testing if the given metalevel is allowed in
     * the referencing (not necessarily declaring; cf. named pointcuts) aspect.
     */
    public Pointcut inline(Hashtable renameEnv, Hashtable typeEnv, Aspect context, int cflowdepth) {
        //get the referencing (i.e. inlining) aspect
        abc.meta.weaving.aspectinfo.Aspect asp = (abc.meta.weaving.aspectinfo.Aspect) context;
        
        //get its meta level and the one of this pointcut
        int aspectLevel = asp.getMetaLevel();        
        int metaLevel = this.level;
        
        //get the position information for error messages
        //if a named pointcut surrounds this one, get its (its the *outermost* surrounding named pointcut)
        //position info, otherwise take the own one
        Position position;
        PointcutRef surroundingNamedPointcut = PointcutRef.getCurrentOuterMostRef();
        if(surroundingNamedPointcut == null) {
            position = getPosition();
        } else {
            position = surroundingNamedPointcut.getPosition();
        }
           
        //generally the metalevel of the pointcut may not be larger than the one of the declaring aspect
        if(metaLevel > aspectLevel) {
            if(aspectLevel < 1) {
                //give a nice error message for non-meta aspects
                new MetaLevelException("meta pointcuts are not allowed in a non-meta aspect. " +
                        "Consider declaring this aspect as a meta aspect.",position);
            } else {
                //or the default message
                new MetaLevelException("meta("+metaLevel+") pointcuts are" +
                        " not allowed in a meta("+ aspectLevel +") aspect.",position);
            }
        } else if(delegate instanceof WithinAdvice && metaLevel == aspectLevel){
            //also, if we have adviceexecution, then the meta level of the pointcut has to be
            //less compared to one of the aspect
            if(aspectLevel < 1) {
                //give a nice error message for non-meta aspects
                new MetaLevelException("adviceexecution is not allowed in a non-meta aspect. " +
                        "Consider declaring this aspect as a meta aspect.",position);
            } else {
                //or the default message
                new MetaLevelException("meta("+metaLevel+") adviceexecution is not allowed" +
                        " in a meta("+ aspectLevel +") aspect.",position);
            }
        }

        
        Pointcut pc = delegate.inline(renameEnv, typeEnv, context, cflowdepth);
        return new MetaTagPointcut(pc, level, getPosition());
    }

    /**
     * Have to make sure that we do not get lost in a nondestructive DNF operation.
     */
    public DNF dnf() {
        Pointcut pc = delegate.dnf().makePointcut(delegate.getPosition());
        return new DNF(new MetaTagPointcut(pc, level, getPosition()));
    }

    /**
     * Return a nice string representation.
     */
    public String toString() {
        return "meta(" + level + ") " + delegate.toString();
    }

    /**
     * We want to return our own position, not the one of the delegate.
     */
    public Position getPosition() {
        return super.getPosition();
    }

    //only delegate methods follow
    
    /**
     * {@inheritDoc}
     */
    public boolean equals(Object obj) {
        return delegate.equals(obj);
    }

    /**
     * {@inheritDoc}
     */
    public void getFreeVars(Set result) {
        delegate.getFreeVars(result);
    }

    /**
     * {@inheritDoc}
     */
    public int hashCode() {
        return delegate.hashCode();
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean unify(Pointcut otherpc, Unification unification) {
        return delegate.unify(otherpc, unification);
    }

    /**
     * {@inheritDoc}
     */
    public void registerSetupAdvice(Aspect context, Hashtable typeEnv) {
        delegate.registerSetupAdvice(context, typeEnv);
    }

}
